/*
 * Copyright (c) 2004-2005  
 *              Martin Reinecke (1), Klaus Dolag (1)
 *               (1) Max-Plank-Institute for Astrophysics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include<iostream>
#include<cmath>
#include<fstream>
#include<algorithm>

#include "arr.h"
#include "cxxutils.h"
#include "paramfile.h"
#include "kernel/bstream.h"
#include "kernel/colour.h"
#include "config/config.h"
#include "utils/colourmap.h"

using namespace std;
using namespace RAYPP;

#define NO_SORTBYVAL
#define NO_REVERSESORT
#define NO_STOP_AFTER_N 5
#define NO_FULLABSROB

const float8 Pi=3.14159265358979323846;

struct PARTICLE
  {
  float4 x,y,z,r,ro,I,T;
  int type;
  };

struct zcmp
  {
  int operator()(const PARTICLE &p1, const PARTICLE &p2)
    {
#ifdef SORTBYVAL
#ifdef REVERSESORT
    return p1.T<p2.T;
#else
    return p1.T>p2.T;
#endif
#else
    return p1.z>p2.z;
#endif
    }
  };

class exptable
  {
  private:
    int nexp;
    float8 maxexptab;
    float8 expfac;
    float8 *exptab;

  public:
    exptable (int nexp_, float8 maxexp_)
      : nexp(nexp_), maxexptab(maxexp_), expfac(nexp/maxexptab),
        exptab(new float8[nexp])
      {
      for (int m=0; m<nexp; ++m)
        exptab[m]=exp(m/expfac);
      }

    float8 operator() (float8 arg) const
      {
      arg *= expfac;
      int iarg=int(arg);
      if (iarg<0) return 1.;
      if (iarg>=nexp-1) return 0.;
      float8 frac=arg-iarg;
      return (1.-frac)*exptab[iarg]+frac*exptab[iarg+1];
      }
  };

int gadget_find_block(bifstream &file,const string &label)
  {
  int4 blocksize=0, blksize;
  char blocklabel[5]={"    "};

  file.rewind();

  while(!file.eof() && blocksize == 0)
    {
    file >> blksize;
    if (blksize != 8)
      {
	cout << "wrong structure in GADGET file: " << blksize << endl;
      exit(1);
      }
    else
      {
      file.get(blocklabel,4);
      blocklabel[4] = 0;
      //cout << "blocklabel: " << blocklabel << endl;
      file >> blocksize;
      //cout << "blocksize: " << blocksize << endl;
      file.skip(4);
      if (label!=blocklabel)
	{
        file.skip(blocksize);
        blocksize=0;
        }
      }
    }
  return(blocksize-8);
  }

int gadget_read_header(bifstream &file, int *npart,double *massarr,
  double &time,double &redshift, int *npartall)
  {
  int blocksize = gadget_find_block (file,"HEAD");
  planck_assert (blocksize>0, "Header block not found");
  file.skip(4);
  file.get(npart,6);
  file.get(massarr,6);
  file >> time >> redshift;
  file.skip(8);
  file.get(npartall,6);

  return blocksize;
  }

int main (int argc, const char ** argv)
  {
const float8 Pi=3.14159265358979323846;
  announce ("splotch");
  planck_assert (argc==2,"usage: splotch <parameter file>");
  paramfile params (argv[1]);

  cout << "initializing" << endl;
  int res = params.find<int>("resolution",200);
  string infilename = params.find<string>("infile");
  string outfile = params.find<string>("outfile");
  string filename;
  int numfiles = params.find<int>("numfiles",1);
  bool doswap = params.find<bool>("swap_endian",true);

  int lable_int_nr = params.find<int>("lable_intensity_nr",2);
  int lable_col_nr = params.find<int>("lable_color_nr",1);
  string lable_int,lable_col;
  switch(lable_int_nr)
    {
    case 1:lable_int="TEMP"; break;
    case 2:lable_int="RHO "; break;
    case 3:lable_int="U   "; break;
    }
  switch(lable_col_nr)
    {
    case 1:lable_col="TEMP"; break;
    case 2:lable_col="RHO "; break;
    case 3:lable_col="U   "; break;
    }
  int colors = params.find<int>("colors",1);
  bool log_int = params.find<bool>("log_intensity",true);
  bool log_col = params.find<bool>("log_color",true);
  float minval_int = params.find<float>("min_int",1024);
  float maxval_int = params.find<float>("max_int",1024);
  float minval_col = params.find<float>("min_col",1024);
  float maxval_col = params.find<float>("max_col",1024);

  bool stars = params.find<bool>("stars",false);
  float star_hsml = params.find<float>("star_hsml",5.0);
  int stars_int_nr = params.find<int>("stars_intensity_nr",2);
  int stars_col_nr = params.find<int>("stars_color_nr",1);
  string stars_int,stars_col;
  switch(stars_int_nr)
    {
    case 1:stars_int="AGE "; break;
    }
  switch(stars_col_nr)
    {
    case 1:stars_col="AGE "; break;
    }
  bool log_stars_int = params.find<bool>("log_stars_intensity",true);
  bool log_stars_col = params.find<bool>("log_stars_color",true);
  float minval_stars_int = params.find<float>("min_stars_int",1024);
  float maxval_stars_int = params.find<float>("max_stars_int",1024);
  float minval_stars_col = params.find<float>("min_stars_col",1024);
  float maxval_stars_col = params.find<float>("max_stars_col",1024);
  
  float8 brightness = params.find<double>("brightness",1.);

  exptable xexp(10000,-20.);

  int4 npart[6],npartall[6];
  float8 massarr[6],time,redshift;

  if(numfiles>1) filename=infilename+"."+dataToString(0);
  else           filename=infilename;
  bifstream infile(filename.c_str(),doswap);
  if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
  gadget_read_header(infile,npart,massarr,time,redshift,npartall);
  if(!stars) npartall[4]=0;
  cout <<  time << " " << redshift << endl;
  infile.close();

  int np=npartall[0];
  int nstar=npartall[4];
  cout << "Allocating space for " << np << " gas particles and " << nstar 
       << " star particles ..." << endl;  
  PARTICLE *p=new PARTICLE[np+nstar];
  float4 mintmp=1e30, maxtmp=-1e30,minint=1e30, maxint=-1e30;
  float4 mintmp_stars=1e30, maxtmp_stars=-1e30,minint_stars=1e30, maxint_stars=-1e30;
  int nskip,npstart=0,nstarstart=np;

  for(int f=0;f<numfiles;f++)
    {
     if(numfiles>1) filename=infilename+"."+dataToString(f);
     else           filename=infilename;
     bifstream infile(filename.c_str(),doswap);
     if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
     gadget_read_header(infile,npart,massarr,time,redshift,npartall);
     if(!stars) npart[4]=0;
     nskip=npart[1]+npart[2]+npart[3];
     cout << "Reading " << npart[0] << " particles + " << npart[4] << " stars " << endl;
     cout << "Distributing at " << npstart << " and " << nstarstart << endl;
     gadget_find_block(infile,"POS ");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       infile >> p[npstart+m].x >> p[npstart+m].y >> p[npstart+m].z;
     if(stars)
       {
       cout << "Skipping " << nskip << " Positions ..." << endl;
       if(nskip>0) infile.skip(nskip*4*3);
       if(npart[4]>0)
         {
         for (int m=0; m<npart[4]; ++m)
           infile >> p[m+nstarstart].x >> p[m+nstarstart].y >> p[m+nstarstart].z;
         }
       }

     for (int m=0; m<npart[0]; ++m)
        p[m+npstart].type=0;

     gadget_find_block(infile,"HSML");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       infile >> p[m+npstart].r;

     if (lable_col_nr == 0)
       {
          for (int m=0; m<npart[0]; ++m)
            p[m+npstart].T=1.0;
       }
     else
       {
         gadget_find_block(infile,lable_col);
         infile.skip(4);
         for (int m=0; m<npart[0]; ++m)
           {
           infile >> p[m+npstart].T;
           if(log_col) p[m+npstart].T = log(p[m+npstart].T);
           mintmp=min(mintmp,p[m+npstart].T);
           maxtmp=max(maxtmp,p[m+npstart].T);
           }
       }

     if (lable_int_nr == 0)
       {
          for (int m=0; m<npart[0]; ++m)
            p[m+npstart].I=1.0;
       }
     else
       {
          gadget_find_block(infile,lable_int);
          infile.skip(4);
          for (int m=0; m<npart[0]; ++m)
            {
            infile >> p[m+npstart].I;
            if(log_int) p[m+npstart].I = log(p[m+npstart].I);
            minint=min(minint,p[m+npstart].I);
            maxint=max(maxint,p[m+npstart].I);
            }
       }

     if(stars)
       {
       if(npart[4]>0)
         {
         for (int m=0; m<npart[4]; ++m)
	   {
             p[m+nstarstart].type=1;
             p[m+nstarstart].T=1;
             p[m+nstarstart].r=star_hsml;
	   }
         if (lable_int_nr == 0)
           {
            for (int m=0; m<npart[4]; ++m)
              p[m+nstarstart].I=1.0;
           }
         else
           {
              gadget_find_block(infile,stars_int);
              infile.skip(4);
              for (int m=0; m<npart[4]; ++m)
                {
                infile >> p[m+nstarstart].I;
                if(log_int) p[m+nstarstart].I = log(p[m+nstarstart].I);
                minint_stars=min(minint_stars,p[m+nstarstart].I);
                maxint_stars=max(maxint_stars,p[m+nstarstart].I);
                }
           }
         if (stars_col_nr == 0)
           {
              for (int m=0; m<npart[0]; ++m)
                p[m+npstart].I=1.0;
           }
         else
           {
             gadget_find_block(infile,stars_col);
             infile.skip(4);
             for (int m=0; m<npart[4]; ++m)
               {
               infile >> p[m+nstarstart].T;
               if(log_stars_col) p[m+nstarstart].T = log(p[m+nstarstart].T);
               mintmp_stars=min(mintmp_stars,p[m+nstarstart].T);
               maxtmp_stars=max(maxtmp_stars,p[m+nstarstart].T);
               }
           }
	 }
       }

     npstart+=npart[0];
     nstarstart+=npart[4];
    }

  if(minval_col!=maxval_col) 
    {
    mintmp=minval_col;
    maxtmp=maxval_col;
    }

  if(minval_int!=maxval_int) 
    {
    minint=minval_int;
    maxint=maxval_int;
    }

  if(minval_stars_col!=maxval_stars_col) 
    {
    mintmp_stars=minval_stars_col;
    maxtmp_stars=maxval_stars_col;
    }

  if(minval_stars_int!=maxval_stars_int) 
    {
    minint_stars=minval_stars_int;
    maxint_stars=maxval_stars_int;
    }

  cout << "Gas particles: " << endl;
  cout << "Color Range:     " << mintmp << " (min) , " << 
                                 maxtmp << " (max) " << endl;
  cout << "Intensity Range: " << minint << " (min) , " << 
                                 maxint << " (max) " << endl;
  cout << "Star particles: " << endl;
  cout << "Color Range:     " << mintmp_stars << " (min) , " << 
                                 maxtmp_stars << " (max) " << endl;
  cout << "Intensity Range: " << minint_stars << " (min) , " << 
                                 maxint_stars << " (max) " << endl;

  for (int m=0; m<np+nstar; ++m)
    {
    if(p[m].type==0)
      {
        if(lable_col_nr != 0) p[m].T = (p[m].T-mintmp)/(maxtmp-mintmp);
        if(lable_int_nr != 0) p[m].I = (p[m].I-minint)/(maxint-minint);
      }
    else
      {
        if(stars_col_nr != 0) p[m].T = (p[m].T-mintmp_stars)/(maxtmp_stars-mintmp_stars);
        if(stars_int_nr != 0) p[m].I = (p[m].I-minint_stars)/(maxint_stars-minint_stars);
      }
    }


  cout << "transforming" << endl;
  VECTOR campos(params.find<double>("camera_x"),
                params.find<double>("camera_y"),
                params.find<double>("camera_z"));
  VECTOR lookat(params.find<double>("lookat_x"),
                params.find<double>("lookat_y"),
                params.find<double>("lookat_z"));
  VECTOR sky(params.find<double>("sky_x",0),
                params.find<double>("sky_y",1),
                params.find<double>("sky_z",0));
  sky.Normalize();
  VECTOR zaxis = (lookat-campos).Norm();
  VECTOR xaxis = Cross (sky,zaxis).Norm();
  VECTOR yaxis = Cross (zaxis,xaxis);
  TRANSFORM trans;
  trans.Make_General_Transform
    (TRANSMAT(xaxis.x,xaxis.y,xaxis.z,
              yaxis.x,yaxis.y,yaxis.z,
              zaxis.x,zaxis.y,zaxis.z,
              0,0,0));
  trans.Invert();
  TRANSFORM trans2;
  trans2.Make_Translation_Transform(-campos);
  trans2.Add_Transform(trans);
  trans=trans2;
  //  trans.Add_Transform(trans2);
  double fov = params.find<double>("fov",45); //in degrees
  double fovfct = tan(fov*0.5*Pi/180.);

  for (int m=0; m<np+nstar; ++m)
    {
    VECTOR v(p[m].x,p[m].y,p[m].z);
    v=trans.TransPoint(v);
    p[m].x=v.x; p[m].y=v.y; p[m].z=v.z;
    }
  for (int m=0; m<np+nstar; ++m)
    {
    float8 xfac=1./(fovfct*p[m].z);
    p[m].x = res*.5 * (p[m].x+fovfct*p[m].z)*xfac;
    p[m].y = res*.5 * (p[m].y+fovfct*p[m].z)*xfac;
    p[m].ro = p[m].r;
    p[m].r = p[m].r *res*.5*xfac;
    }

#ifdef SORTBYVAL
#ifdef REVERSESORT
  cout << "reverse sorting by value" << endl;
#else
  cout << "sorting by value" << endl;
#endif
#else
  cout << "sorting by z" << endl;
#endif
  sort(&p[0], &p[np+nstar], zcmp());

  cout << "plotting" << endl;

  float8 rfac=1.5;
  float8 powtmp = pow(float8(Pi/2),float8(1./3.));
  float8 sigma0=powtmp/sqrt(2*Pi);
  float8 i00=brightness/(2*sqrt(Pi)*powtmp);

  arr2<COLOUR> pic(res,res);
#ifdef STOP_AFTER_N
  int z_count[res][res];
#endif

  COLOURMAP amap,emap;

    COLOUR c0,c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15;
    c0=COLOUR(0,0,0);
    c1=COLOUR(0,0,1);
    c2=COLOUR(0,1,0);
    c3=COLOUR(0,1,1);
    c4=COLOUR(1,0,0);
    c5=COLOUR(1,0,1);
    c6=COLOUR(1,1,0);
    c7=COLOUR(0.66,0.66,0.66);
    c8=COLOUR(0.33,0.33,0.33);
    c9=COLOUR(0,0,0.5);
    c10=COLOUR(0,0.5,0);
    c11=COLOUR(0,0.5,0.5);
    c12=COLOUR(0.5,0,0);
    c13=COLOUR(0.5,0,0.5);
    c14=COLOUR(0.5,0.5,0);
    c15=COLOUR(1,1,1);

  switch(colors)
    {
    case 0:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY( 0,.5,c4,c7));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,1.,c7,c1));
      emap=amap;
      break;
    case 1:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY( 0,.2,COLOUR(1,0,0),COLOUR(0,1,0)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.2,.5,COLOUR(0,1,0),COLOUR(0,0,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,.65,COLOUR(0,0,1),COLOUR(0,1,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.65,.8,COLOUR(0,1,1),COLOUR(1,0,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.8,1.,COLOUR(1,0,1),COLOUR(1,1,1)));
      emap=amap;
      break;
    case 2:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.33,COLOUR(0.00,0.00,1.00),COLOUR(1.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(1.00,0.66,0.66),COLOUR(0.50,1.00,0.50)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,1.00,0.50),COLOUR(0.00,0.66,0.33)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.33),COLOUR(1.00,0.00,0.00)));
      emap=amap;
      break;
    case 3:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.0,.5,COLOUR(0.0,0.0,1.0),COLOUR(0.5,1.0,0.5)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,1.,COLOUR(0.5,1.0,0.5),COLOUR(1.0,0.0,0.0)));
      emap=amap;
      break;
    case 4:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.16,COLOUR(0.00,0.00,0.00),COLOUR(0.00,0.00,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.16,.33,COLOUR(0.00,0.00,1.00),COLOUR(1.00,0.66,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(1.00,0.66,1.00),COLOUR(0.50,1.00,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,1.00,1.00),COLOUR(0.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.66),COLOUR(1.00,0.00,0.00)));
      emap=amap;
      break;
    case 5:
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.14,c0,c9));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.14,.29,c9,c1));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.29,.43,c1,c4));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.43,.57,c4,c2));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.57,.71,c2,c6));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.71,.86,c6,c15));
      //        amap.Add_Entry(new LINEAR_CMAP_ENTRY(.86,1.0,c,c));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.16,c0,c9));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.16,.33,c9,c1));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,c1,c4));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,c4,c2));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,.83,c2,c6));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.83,1.0,c6,c15));
      emap=amap;
      break;
    case 6:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.07,c0,c9));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.07,.13,c9,c1));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.13,.20,c1,c12));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.20,.27,c12,c13));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.27,.33,c13,c10));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.40,c10,c4));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.40,.47,c4,c8));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.47,.53,c8,c11));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.53,.60,c11,c5));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.60,.67,c5,c14));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.67,.73,c14,c2));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.73,.80,c2,c7));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.80,.87,c7,c3));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.87,.93,c3,c6));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.93,1.0,c6,c15));
      emap=amap;
      break;
    }

  bool gas = params.find<bool>("gas",true);
  float zmaxval = params.find<float>("zmax",1.e23);
  float zminval = params.find<float>("zmin",0.0);
  float grayabsorb = params.find<float>("gray_absorbtion",0.2);

  for (int m=0; m<np+nstar; ++m)
    {
    if ((m%100000)==0)
      cout << m << " of " << np+nstar << " Particles ..." << endl;

    if (p[m].z<=0) continue;
    if (p[m].z<=zminval) continue;
    if (p[m].z>=zmaxval) continue;
    //    cout << zminval << " " << p[m].z << " " << zmaxval << endl;
    if (!gas and p[m].type==0) continue;
    if (!stars and p[m].type==1) continue;
#ifdef STOP_AFTER_N
    int pxt=int(p[m].x),pyt=int(p[m].y);
    if (pxt < 0 || pxt >= res) continue;  
    if (pyt < 0 || pyt >= res) continue;  
    z_count[pxt][pyt]++;
    if (z_count[pxt][pyt] > STOP_AFTER_N) continue;
#endif

    float8 temp=p[m].T;
    temp=max(float8(0.),min(float8(1.),temp));
    float8 intensity=p[m].I;
    intensity=max(float8(0.),min(float8(1.),intensity));
    COLOUR a,e;
    //    if(p[m].type==0)
    //      {
      e=amap.Get_Colour(temp)*intensity;
#ifdef FULLABSROB
      a=COLOUR(1,1,1);
#else
      a=emap.Get_Colour(temp)*intensity;
#endif
      //      }
      //    else
      //      {
      //      a=COLOUR(1,1,0);
      //      e=COLOUR(1,1,0);
      //      }

    COLOUR q (e.r/(a.r+grayabsorb),e.g/(a.g+grayabsorb),e.b/(a.b+grayabsorb));

    float8 r=p[m].r, ro=p[m].ro;
    float8 sigma=ro*sigma0;
    float8 fac0=-0.5/(sigma*sigma);
    float8 i0=i00/ro;
    float8 radsq = rfac*rfac*r*r;
    float8 prefac1 = ro/r*ro/r*fac0;
    float8 prefac2 = -0.5*i0;

    float8 posx=p[m].x, posy=p[m].y;
    int minx = max(int(posx-rfac*r),0), maxx = min(int(posx+rfac*r+1),res),
        miny = max(int(posy-rfac*r),0), maxy = min(int(posy+rfac*r+1),res);
    for (int y=miny; y<maxy; ++y)
      {
      float8 ysq=(y-posy)*(y-posy);
      for (int x=minx; x<maxx; ++x)
        {
        float8 dsq = (x-posx)*(x-posx) + ysq;
        if (dsq<radsq)
          {
          float8 fac = prefac2*xexp(prefac1*dsq);
          pic[x][y].r = q.r+(pic[x][y].r-q.r)*xexp(fac*a.r);
          pic[x][y].g = q.g+(pic[x][y].g-q.g)*xexp(fac*a.g);
          pic[x][y].b = q.b+(pic[x][y].b-q.b)*xexp(fac*a.b);
          }
        }
      }
    }

  cout << "writing" << endl;
  const byte header[18] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    res%256, res/256, res%256, res/256, 24, 32 };

  bofstream file(outfile.c_str(),file_is_natural);
  for (int m=0; m<18; ++m) file << header[m];
  for (int y=0; y<res; ++y)
    {
    for (int x=0; x<res; ++x)
      {
      pic[x][y].b=min(float8(1.), max(float8(0.), float8(pic[x][y].b)));
      byte pix = min(byte(255),byte(256*pic[x][y].b));
      file << pix;
      pic[x][y].g=min(float8(1.), max(float8(0.), float8(pic[x][y].g)));
      pix = min(byte(255),byte(256*pic[x][y].g));
      file << pix;
      pic[x][y].r=min(float8(1.), max(float8(0.), float8(pic[x][y].r)));
      pix = min(byte(255),byte(256*pic[x][y].r));
      file << pix;
      }
    }
  }
